;#define pic54
#define pic84

#ifdef pic84
	LIST P=16F84, F=INHX8M, R=DEC
#else
	LIST P=16C54, F=INHX8M, R=DEC
#endif

;RPM03.ASM
;
;Digital tachometer using pulse count in fixed gate period
;RPM displayed as RPM/100 on 2 digit 7 segment multiplexed
;display.
;Bargraph capability added.
;4 multiplexed digits/arrays driven through NPN transistors.

;18 JAN 96: If overflow make sure that bargraph stays full on
;	    by setting bar_high, bar_low = F,F
;
;16 JAN 96: Copied from RPM02.ASM
;	    Give overrange indication on digits with "o/r"
;	    on display
;
;19 DEC 95: Segments 'd' and 'g' for LSD swapped round.
;	    Swap seg_72/73 calls to correct bargraph direction.
;
;18 DEC 95: Correct the digit display bit patterns.
;	    For PIC84 set b4 of RA as an input.
;
;15 DEC 95: Allow for programming 84,54 types since
;	    they have different reset vectors
;
;14 DEC 95: store character patterns in blocks of 16 bytes,
;	    check that number is 0-15 before calling the
;	    translation table.
;
;05 DEC 95: revise bit pattern for each digit - no longer the
;	    same between all of them
;	    digit drivers as external Tr -> polarity reversal
;	    for digit driver definitions (no current restriction)
;
;11 NOV 95: rearrange code so sub-routines are near the start
;
;10 NOV 95: add the capacity for a bargraph display,
;           all LEDs low current so can use 1 pin to sink
;           current from a multiplexed set of LEDs
;
;09 OCT 95: change the MSD, LSD allocation on RA0-3 to match
;           that used on the PCB - should've done this on 4/10
;
;06 OCT 95: change the WDT prescaler to 1:32 so its closer
;           to the cylce time
;
;04 OCT 95: change bit pattern used to address segments
;           in order to match that used on PCB layout
;
;22 SEP 95: leading zero blanking
;           change code for 'F' to blank digit
;
;09 SEP 95: more comprehensive commenting
;
;30 AUG 95: initial coding
;
;Simple RPM measurement on pulse count/fixed gate period
;output is scaled to read pulses as RPMx100
;Uses RTCC to count the number of incoming pulses over
;a fixed time - determined by software delay loop.
;Delay loop used to drive two 7 segment displays as MPX.
;
;
;All timings based on 4MHz clock
;
;Common cathode display.
;For each LED digit use two pins to sink digit drive current
;e.g. digit 0 - RA0,1 and digit 1 - RA2,3.
;!05/12/95      RA0 = lsd digit
;               RA1 = msd digit
;               RA2 = upper/right 8 bargraph
;               RA3 = lower/left  8 bargraph

;***INPUTS***
;RTCC = i/p from ignition

;***OUTPUTS***
;PA0 = common cathode sinks for lsd
;PA1 = common cathode sinks for msd
;PA2 = common cathode sinks for bargraph(1)
;PA3 = common cathode sinks for bargraph(2)

;PB0-7 = segment drivers for LEDs
;bit pattern allocation for segments:
;--------------
;RB-> 01234567
;RA|0 egdcPbaf		0 edgcPbaf
;   1 Pcdefgab
;   2 abcdefgh
;   3 hgfedcba
;--------------

CA      SET       1     ;1 = common anode
			;use bit pattern for PCB
			;0 = common cathode
			;use straight thru pattern
;Register allocation/ labels
;
INDRCT  EQU       0    ;For indirect addressing
PC      EQU       2    ;program counter
STATUS  EQU       3    ;status word register

CY      EQU       0    ;carry bit in STATUS
Z       EQU       2    ;zero bit in STATUS

FSR     EQU       4    ;File select register
portA   EQU       5    ;Register allocation
portB   EQU       6

RTCC    EQU       1    ;8 bit counter

cycles  EQU       8    ;nr. of cycles to output
delay   EQU       9    ;delay counter to produce 12uS delay





limit   EQU       11   ;time-out value for receiver stored here

R       EQU       16    ;holds nr. for BCD conversion
			;=nr of pulses received by RTCC (i.e. RPM)

;BCD result numbers
K0      EQU       17    ;BCD units
K1      EQU       18    ;BCD tens
K2      EQU       19    ;BCD hundreds

;bar graph numbers
BR      EQU     20      ;holds a copy of R for the bargraph
bar_graph       EQU     21

BAR0    EQU     22
bar_low	EQU	22

BAR1    EQU     23
bar_high	EQU	23

;-----------------------------------------------------------------
;Literals (constants/reference values)
;

DISP_LOOPS      EQU     0x60	;outside loop time = 300mS

DEL_5MS         EQU     0       ;=768uS delay

FINE_TUNE       EQU     1	;

BAR_MID		EQU	9	;mid point of the bargraph

;bit patterns for digit drivers
LSD_DIGIT	EQU     b'00000001'     ;LSD
MSD_DIGIT	EQU     b'00000010'     ;MSD
MSD_BAR		EQU     b'00000100'     ;bar(2)
LSD_BAR		EQU     b'00001000'     ;bar(1)
ALLOFF		EQU     b'00000000'	;all off

;-----------------------------------------------------------------
;subroutines go here near the start of the page (0x00)
	org     0x10
;divide by 5 for bargraph
;done by successive subtractions
div_5
	movlw   0xff
	movwf   bar_graph
suc_sub
	incf    bar_graph
	movlw   D'5'
	subwf   BR              ;BR=BR-5
	btfsc   STATUS,CY
	goto    suc_sub
;now bar_graph = BR/5
;convert the value in bar_graph to bit patterns in BAR0, BAR1
	movlw	BAR_MID
	subwf	bar_graph,W	;W=bar_graph-BAR_MID
	btfsc	STATUS,CY
	goto	is_over_BAR_MID

	movfw	bar_graph
	movwf	bar_low		;bar_low=bar_graph
	clrf	bar_high	;bar-high=0
	goto	bar_end

is_over_BAR_MID
	movwf	bar_high
	incf	bar_high	;bar_high=bar_graph-BAR_MID+1
	movlw	8
	movwf	bar_low		;bar_low=8

bar_end
	retlw   0
;end of sub-routine

;8 bit binary to 3 digit BCD conversion
;R = binary value
;K0-2 = BCD values
;works by successive subtraction
bin_bcd 
	CLRF    K0      ;units
	CLRF    K1      ;tens
	CLRF    K2      ;hundreds

do_hund 
	MOVLW   D'100'
	SUBWF   R,W             ;W=R-100
	BTFSS   STATUS,CY       ;R<100?
	GOTO    do_ten          ;yes, jump to tens
	INCF    K2              ;K2->K2+1
	MOVWF   R               ;and R->R-100
	GOTO    do_hund         ;repeat loop

do_ten  
	MOVLW   D'10'
	SUBWF   R,W             ;W=R-10
	BTFSS   STATUS,CY       ;R<10?
	GOTO    do_unit         ;yes, jump to units
	INCF    K1              ;K1->K1+1
	MOVWF   R               ;and R->R-10
	GOTO    do_ten          ;repeat loop

do_unit 
	MOVFW   R
	MOVWF   K0              ;store R in K0

;22 SEP 95: check if k2,k1 are zero -> make='F'
;so that leading zero is blanked.
;16 JAN 95: overrange indication with 'E' code

	MOVF    K2              ;K2=K2, affects Z flag
	SKPNZ                   ;skip if K2!=0
	GOTO    chkK1           ;go if K2!=0
	MOVLW   H'0E'           ;'E'=code for o/r
	MOVWF   K2
	MOVWF   K1
	MOVWF   K0
;also set bar_high and bar_low to F,F so that bargraph
;is all ON
	MOVLW	H'0F'
	MOVWF	bar_high
	MOVWF	bar_low
	GOTO	endbcd
chkK1
	MOVF    K1              ;as above but for K1
	SKPZ                    ;skip if K1=0
	GOTO    endbcd
	MOVLW   H'0F'
	MOVWF   K1
				;K0 cannot be blanked out!

endbcd  
	RETLW   0

;end of bin_bcd

;in W = BCD value
;returns W = segment pattern
;common cathode - 0.5"
;bit pattern allocation to segments:
;b'76543210'
;b'fabPcgde'
seg_70   
;for digit 0 = lsd
	ADDWF   PC
	RETLW   b'00111111'     ;0
	RETLW   b'00000110'     ;1
	RETLW   b'01011011'     ;2
	RETLW   b'01001111'     ;3
	RETLW   b'01100110'     ;4
	RETLW   b'01101101'     ;5
	RETLW   b'01111101'     ;6
	RETLW   b'00000111'     ;7
	RETLW   b'01111111'     ;8
	RETLW   b'01101111'     ;9
	RETLW   b'00000000'     ;A=blank
	RETLW   b'00000000'     ;B=blank
	RETLW   b'00000000'     ;C=blank
	RETLW   b'00000000'     ;D=blank
	RETLW   b'01010000'     ;E=r of o/r
	RETLW   b'00000000'     ;F=blank
;end of 7 seg pattern for D0

;b'bagfedcP'
;b'76543210'
seg_71   
;for digit 1 = MSD
	ADDWF   PC
	RETLW   b'10111111'     ;0
	RETLW   b'10000110'     ;1
	RETLW   b'11011011'     ;2
	RETLW   b'11001111'     ;3
	RETLW   b'11100110'     ;4
	RETLW   b'11101101'     ;5
	RETLW   b'11111101'     ;6
	RETLW   b'10000111'     ;7
	RETLW   b'11111111'     ;8
	RETLW   b'11101111'     ;9
	RETLW   b'10000000'     ;A=blank
	RETLW   b'10000000'     ;B=blank
	RETLW   b'10000000'     ;C=blank
	RETLW   b'10000000'     ;D=blank
	RETLW   b'01111001'     ;E=o of o/r
	RETLW   b'00000000'     ;F=blank
;end of 7 seg pattern for D1

seg_72  
;for array 0
	ADDWF   PC
	RETLW   b'00000000'     ;0
	RETLW   b'00000001'     ;1
	RETLW   b'00000011'     ;2
	RETLW   b'00000111'     ;3
	RETLW   b'00001111'     ;4
	RETLW   b'00011111'     ;5
	RETLW   b'00111111'     ;6
	RETLW   b'00111111'     ;7
	RETLW   b'00111111'     ;8
	RETLW   b'00111111'     ;9
	RETLW   b'00111111'     ;A
	RETLW   b'00111111'     ;B
	RETLW   b'00111111'     ;C
	RETLW   b'00111111'     ;D
	RETLW   b'00111111'     ;E
	RETLW   b'00111111'     ;F

;end of 7 seg pattern for D2

seg_73
;for array 1
	ADDWF   PC
	RETLW   b'00000000'     ;0
	RETLW   b'10000000'     ;1
	RETLW   b'11000000'     ;2
	RETLW   b'11100000'     ;3
	RETLW   b'11110000'     ;4
	RETLW   b'11111000'     ;5
	RETLW   b'11111100'     ;6
	RETLW   b'11111110'     ;7
	RETLW   b'11111111'     ;8
	RETLW   b'11111111'     ;9
	RETLW   b'11111111'     ;A
	RETLW   b'11111111'     ;B
	RETLW   b'11111111'     ;C
	RETLW   b'11111111'     ;D
	RETLW   b'11111111'     ;E
	RETLW   b'11111111'     ;F

;end of 7 seg pattern for D3
;*****************************************************************
;-----------------------------------------------------------------
;main program here at the middle of the page (0x100)
	org     0x100

start   CLRF    R               ;zero rpm count when first started

;the main loop is here
forever 
#ifdef pic84
	MOVLW   b'00010000'     ;portA 0-3 is o/p, 4 is i/p(RTCC)
#else
	MOVLW   b'00000000'     ;portA is all o/p
#endif
	TRIS    portA           ;set i/o state

	MOVLW   b'00000000'     ;portB, b0-b7=o/p
	TRIS    portB           ;set i/o state
;reassign the i/o ports on each cycle in case they are
;corrupted for whatever reason
	
	CLRWDT                  ;clear watchdog
;watchdog is cleared, so need to re-assign the prescaler
	MOVLW   b'00101101'     ;prescaler 1:32, to WDT, ext RTCC clk+leading edge
	OPTION                  ;assign it now

	MOVFW   R
	MOVWF   BR              ;place a copy of R in BR
	CALL    div_5           ;convert BR to bar in BAR0, BAR1
	CALL    bin_bcd         ;convert R to BCD in K0, K1, K2
;new timing cycle starts here
	CLRF    RTCC            ;reset the counter
				
;-----------------------------------------------------------------
;
;this section lasts for 300mS gate period 
;after which read the RTCC and store in R
;during that time the last stored pulse count
;is displayed bt multiplexing the LEDs, 5mS /digit


	MOVLW   DISP_LOOPS      ;outside loop to get 300mS delay
	MOVWF   cycles          ;store in cycle counter

stay    
;put up LSD on LED0 for display first
	MOVFW   K0
	ANDLW   H'0f'           ;mask msb's
	CALL    seg_70          ;convert K0 to 7 seg pattern
	MOVWF   portB           ;output K0 pattern = lsd
	MOVLW   LSD_DIGIT       ;set bit pattern and ...
	MOVWF   portA           ;... drive LED0 cathode low

;now a delay for 5mS to allow the LSD to be seen
	MOVLW   DEL_5MS
	MOVWF   delay
del_1   
	DECFSZ  delay
	GOTO    del_1

	MOVLW   ALLOFF
	MOVWF   portA           ;both digits OFF

;put up MSD on LED1 for display
	MOVFW   K1
	ANDLW   h'0f'           ;mask msb's
	CALL    seg_71          ;convert K1 to 7 seg pattern

	MOVWF   portB           ;output K1 pattern = msd
	MOVLW   MSD_DIGIT       ;set bit pattern and ...
	MOVWF   portA           ;...drive LED1 cathode low

;now a delay for 5mS to allow the MSD to be seen
	MOVLW   DEL_5MS
	MOVWF   delay
del_2   
	DECFSZ  delay
	GOTO    del_2

	MOVLW   ALLOFF
	MOVWF   portA           ;both digits OFF

;put up lower 8 LEDs of bargraph
	MOVFW   BAR0            ;get the bar value
	ANDLW   h'0f'           ;mask msb's
	call	seg_72		;convert to pattern

	MOVWF   portB		;output the pattern
	MOVLW   LSD_BAR
	MOVWF   portA           ;light the LEDs

;now a delay for 5mS to allow the MSD to be seen
	MOVLW   DEL_5MS
	MOVWF   delay
del_3   
	DECFSZ  delay
	GOTO    del_3

	MOVLW   ALLOFF
	MOVWF   portA           ;both digits OFF

;put up upper 8 LEDs of bargraph
	MOVFW   BAR1            ;get the bar pattern
	ANDLW   h'0f'           ;mask msb's
	call	seg_73		;convert to pattern

	MOVWF   portB
	MOVLW   MSD_BAR
	MOVWF   portA           ;light the LEDs

;now a delay for 5mS to allow the MSD to be seen
	MOVLW   DEL_5MS
	MOVWF   delay
del_4   
	DECFSZ  delay
	GOTO    del_4

;end of display MPX
	MOVLW   ALLOFF
	MOVWF   portA           ;both digits OFF

;any fine tuning for gate period goes here
	

	MOVLW   FINE_TUNE       ;these three steps take
	MOVWF   delay           ;3uS -> loop lasts 1570uS
	MOVFW   delay
	MOVLW   FINE_TUNE


;...any extra delay to improve accuracy...
	DECFSZ  cycles
	GOTO    stay

;any extra fine tuning for gate period goes here

;...any extra delay to improve accuracy...

;gate period over
;-----------------------------------------------------------------
;now read RTCC -> R for display on next cycle
	
	
	MOVFW   RTCC
	MOVWF   R

;repeat the display/measurement cycle forever
	GOTO    forever

;-----------------------------------------------------------------

#ifdef pic84
	ORG     0h
#else
	ORG     1FFh
#endif
	GOTO    start

	END
